﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace SC
{
    public class MyCSVReader
    {
        #region メンバ変数
        /// <summary>
        /// ダブルクォーテーションです。
        /// </summary>
        private char _dq = '\"';

        /// <summary>
        /// デリミタ(区切り文字)です。
        /// </summary>
        private char _delimiter = ',';

        /// <summary>
        /// CSVファイルパスです。
        /// </summary>
        private string _filepath = string.Empty;

        /// <summary>
        /// CSVファイル読み込み時の文字コードです。
        /// </summary>
        private Encoding _encoding;

        /// <summary>
        /// CSVデータです。
        /// </summary>
        private List<string[]> _data = new List<string[]>();

        /// <summary>
        /// エラー情報です。
        /// </summary>
        private List<string> _errorInfo = new List<string>();
        #endregion

        #region コンストラクタ
        /// <summary>
        /// コンストラクタです。
        /// </summary>
        /// <param name="filePath">CSVファイル名</param>
        /// <param name="encoding">文字コード</param>
        /// <param name="delimiter">デリミタ</param>
        public MyCSVReader(string filePath, Encoding encoding, char delimiter)
        {
            _filepath = filePath;
            _encoding = encoding;
            _delimiter = delimiter;
        }
        #endregion

        #region プロパティ
        /// <summary>
        /// CSVデータを返します。
        /// </summary>
        public List<string[]> Data
        {
            get { return _data; }
        }
        #endregion

        /// <summary>
        /// CSVファイルを読み込み、デリミタで区切った文字列をDataプロパティに設定していきます。
        /// </summary>
        /// <returns>
        /// 0  :正常終了
        /// -1 :ファイルが見つからない
        /// -2 :その他エラー
        /// </returns>
        public int Read()
        {
            try
            {
                string line = string.Empty;
                int offset = 0;

                if (!System.IO.File.Exists(_filepath))
                {
                    _errorInfo.Add(_filepath + "が見つかりません。");
                    return -1;
                }

                // CSVファイルのテキストを全て読み込みます。
                string text = ReadFile(_filepath, _encoding);

                // 1行ずつ項目を格納します。
                while (offset != -1)
                {
                    line = string.Empty;
                    offset = ReadLine(ref line, text, offset);
                    if (0 < line.Trim().Length)
                    {
                        string[] items = Parse(line, _dq, _delimiter);
                        _data.Add(items);
                    }
                }

                return 0;
            }
            catch (System.Exception ex)
            {
                _errorInfo.Add(ex.Message);
                return -2;
            }
        }

        /// <summary>
        /// 指定したファイルから現在位置から末尾までを読み込みます。
        /// </summary>
        /// <param name="path">CSVファイルパス</param>
        /// <param name="encoding">文字コードです。</param>
        /// <returns>ファイルの内容</returns>
        private string ReadFile(string path, Encoding encoding)
        {
            string text = string.Empty;
            using (FileStream fs = new FileStream(path, FileMode.Open, FileAccess.Read, FileShare.ReadWrite))
            {
                StreamReader sr = new StreamReader(fs, encoding);
                text = sr.ReadToEnd();
            }
            return text;
        }

        /// <summary>
        /// 指定したオフセットから最初に見つかった改行コードまでの文字列を返します。
        /// </summary>
        /// <param name="line">取得した文字列</param>
        /// <param name="text">ファイルの内容</param>
        /// <param name="offset">オフセット</param>
        /// <returns>オフセットです。-1の場合はファイルの終端までたどり着いたことを表します。</returns>
        private int ReadLine(ref string line, string text, int offset)
        {
            bool dq = false;

            for (int i = offset; i < text.Length; i++)
            {
                if (!dq && text[i] == '\r')
                {
                    continue;
                }
                if (!dq && text[i] == '\n')
                {
                    return i + 1;
                }

                line += text[i];

                if (text[i] == _dq)
                {
                    if (!dq) { dq = true; }
                    else { dq = false; }
                }
            }
            return -1;
        }

        /// <summary>
        /// 文字列をデリミタで分割した結果を返します。
        /// </summary>
        /// <param name="text">一行分のCSVデータ</param>
        /// <returns>分割結果</returns>
        private string[] Parse(string text, char dq, char delimiter)
        {
            bool dqStart = false;
            bool dqFlg = false;
            string item = string.Empty;
            List<string> items = new List<string>();

            if (string.IsNullOrEmpty(text)) { return null; }

            foreach (char c in text)
            {
                if (c == dq)
                {
                    if (!dqStart && item.Trim().Length == 0)
                    {
                        item = string.Empty;
                        dqStart = true;
                        continue;
                    }
                    if (!dqFlg)
                    {
                        dqFlg = true;
                        continue;
                    }

                    item += dq;
                    dqFlg = false;
                }
                else if (c == delimiter)
                {
                    if (dqStart && !dqFlg)
                    {
                        item += c;
                        continue;
                    }
                    items.Add(item);
                    item = string.Empty;
                    dqStart = false;
                    dqFlg = false;
                }
                else
                {
                    if (dqStart && dqFlg) { continue; }
                    item += c;
                }
            }
            items.Add(item);

            return items.ToArray();
        }
    }
}
